# itunesdb.tcl
#
# Tcl library for reading iTunesDB files.
#
# Copyright (C) 2004 Andy Goth <unununium@openverse.com>
# For more information visit <URL:http://ioioio.net/devel/itunesdb/>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# See the file README for information on how to get the full text of the GNU
# General Public License.

package require snit     0.97
package provide itunesdb 0.02

snit::type ::itunesdb_impl {
    pragma -hastypemethods no
    pragma -simpledispatch yes

    variable path       ;# Name of iTunesDB file.
    variable chan       ;# Tcl channel for reading the file.

    variable state      ;# What is currently being parsed?
    variable index      ;# Next track or playlist to read.
    variable pos        ;# Current file position in bytes.

    variable track_pos  ;# File position at which first track resides.
    variable track_num  ;# Total number of tracks in the file.

    variable plist_pos  ;# File position at which first playlist resides.
    variable plist_num  ;# Total number of playlists in the file.

    # itunesdb $filename
    #
    # Opens an iTunesDB database file.
    #
    # $filename - Path to the iTunesDB file.
    #
    # return - Command through which further access is conducted ($db).
    constructor {filename} {
        set path $filename
        set chan [open $path r]

        # Treat the iTunesDB file as a raw binary--- no translations.
        fconfigure $chan -translation binary

        set pos   0         ;# Start at the beginning of the file.
        set state initial   ;# Initial state: expect 'mhbd'.

        # Get file summary information.
        $self read
    }

    # $db destroy
    #
    # Closes the database file.
    destructor {
        if {[info exists chan]} {
            close $chan
        }
    }

    # $db count $which
    #
    # Returns the number of elements of the given type in the database file.
    #
    # $which - Element type to count.  Can be track or playlist.
    #
    # return - Number of tracks or playlists in the file.
    method count {which} {
        switch -- $which {
        track {
            return $track_num
        } playlist {
            return $plist_num
        } default {
            error "bad type \"$which\": must be track or playlist"
        }}
    }

    # $db seek $which
    #
    # Changes the current position in the database file.
    #
    # $which - Next section to read from.  Can be track or playlist.
    #
    # return - A list of the format {$section $index}, as in [$db tell].
    method seek {which} {
        switch -- $which {
        track {
            set num   $track_num
            set pos   $track_pos
            set state track_data
        } playlist {
            set num   $plist_num
            set pos   $plist_pos
            set state plist_data
        } default {
            error "bad type \"$which\": must be track or playlist"
        }}

        set index 0
        if {$num == 0} {
            set    state end
            return [list end    0]
        } else {
            return [list $which 0]
        }
    }

    # $db tell
    #
    # Returns the current position in the database file.
    #
    # return - A list of the format {$section $index}.
    #     $section - Current section type: other, end, track, or playlist.
    #     $index   - Index of the next record in this section.
    method tell {} {
        switch -- $state {
        initial    {return [list other    0     ]}
        end        {return [list end      0     ]}
        track_data {return [list track    $index]}
        plist_data {return [list playlist $index]}
        default    {error "BUG: invalid state \"$state\""}
        }
    }

    # $db remaining
    #
    # Returns the number of records remaining.
    #
    # return - Number of records.
    method remaining {} {
        switch -- $state {
        initial    {return 0}
        end        {return 0}
        track_data {return [expr {$track_num - $index}]}
        plist_data {return [expr {$plist_num - $index}]}
        default    {error "BUG: invalid state \"$state\""}
        }
    }

    # $db read
    #
    # Reads and returns the next record in the database file.
    #
    # return - The data read from the file.  The format depends on the most
    #     recent call to [$db seek].
    #     (initial) - {summary $filename track_count $num playlist_count $num}
    #     track     - {track $id key $val key $val ...}
    #     playlist  - {playlist $name type $type tracks {$id $id $id ...}}
    # error - "file is not an iTunes DB" - Incorrect format on first read.
    #     "no more records" - Attempt to read past end of section.
    method read {} {
        set next_pos $pos

        switch -- $state {
        initial    {set goal summary }
        track_data {set goal track   }
        plist_data {set goal playlist} 
        end        {error "no more records"              }
        default    {error "BUG: invalid state \"$state\""}
        }

        while {1} {
            # Read all bytes for this block.  Get the header.
            if {[catch {
                seek $chan $pos
                set buffer [read $chan 12]
                set header [GetHdr $buffer 0]

                if {$header eq "mhod" || $header eq "mhip"} {
                    set size [GetInt $buffer 8]
                } else {
                    set size [GetInt $buffer 4]
                }

                append buffer [read $chan [expr {$size - 12}]]
            } err]} {
                set state end
                if {[eof $chan]} {
                    if {$state eq "initial"} {
                        error "file is not an iTunes DB"
                    } else {
                        break
                    }
                } else {
                    error $err
                }
            }

            switch -- [catch {

            # Process this block.
            switch -glob -- "$state $header" {
            "initial mhbd" {
                # Found initial header.  Skip ahead to mhsd.
                incr next_pos  [GetInt $buffer 4]
                set  state     ready
            } "initial *" {
                # Bad bits at the beginning of the file.
                set  err_pos   0
                set  err_state end
                error "file is not an iTunes DB"
            } "ready mhsd" {
                # Top-level aggregate block.
                incr next_pos  [GetInt $buffer 4]
                set  next_mhsd [expr {$pos + [GetInt $buffer 8]}]
                set  err_state ready
                set  err_pos   $next_mhsd
                set  type      [GetInt $buffer 12]
                switch -- $type {
                1 {
                    # Track list.
                    if {[info exists track_pos]} {
                        error "extra track list section"
                    }
                    set state track_header
                } 2 {
                    # Playlists.
                    if {[info exists plist_pos]} {
                        error "extra playlist section"
                    }
                    set state plist_header
                } default {
                    # Neither track list nor playlist.
                    set next_pos $next_mhsd
                }}
            } "track_header mhlt" {
                # Beginning of track list.
                set track_pos [expr {$pos + [GetInt $buffer 4]}]
                set track_num [GetInt $buffer 8]
                set next_pos  $next_mhsd
                set state     ready
            } "plist_header mhlp" {
                # Beginning of playlists.
                set plist_pos [expr {$pos + [GetInt $buffer 4]}]
                set plist_num [GetInt $buffer 8]
                set next_pos  $next_mhsd
                set state     ready
            }
                
            "track_data mhit" {
                # Track entry, numeric portion.
                incr next_pos [GetInt $buffer 4]
                set  err_pos  [expr {$pos + [GetInt $buffer 8]}]

                if {$index == $track_num - 1} {
                    set err_state end
                } else {
                    set err_state track_data
                }

                set fld_num [GetInt $buffer 12]
                set result  [list track [GetInt $buffer 16]]

                foreach {field offset} {
                    rating     28     last_mod  32     file_size   36
                    duration   40     track_num 44     track_count 48
                    year       52     bit_rate  56     volume      64
                    play_count 80     last_play 88     cd_num      92
                    cd_count   96
                } {
                    set val [GetInt $buffer $offset]

                    switch -- $field {
                    rating {
                        set val [expr {$val >> 24 / 20}]
                    } last_mod - last_play {
                        set val [expr {$val == 0 ? "never" : $val - 2082844800}]
                    }}

                    lappend result $field $val
                }

                if {$fld_num == 0} {
                    # No text fields to read, so quit now.
                    break
                } else {
                    # Expect text fields (mhod).
                    set fld_idx 0
                    set state   track_data_text
                }
            } "track_data_text mhod" {
                # Track entry, textual portion.
                incr next_pos [GetInt $buffer 8]

                # Determine field type.
                switch -- [GetInt $buffer 12] {
                 1 {set field title    }       2 {set field ipod_path}
                 3 {set field album    }       4 {set field artist   }
                 5 {set field genre    }       6 {set field type     }
                 7 {set field equalizer}       8 {set field comment  }
                12 {set field composer } default {set field unknown  }
                }

                if {$field ne "unknown"} {
                    # Read textual data.
                    lappend result $field
                    lappend result [GetUtf16 $buffer 40 [GetInt $buffer 28]]
                }

                incr fld_idx
                if {$fld_idx == $fld_num} {
                    # No more text fields to read.  Quit and return the record.
                    break
                }
            }
                
            "plist_data mhyp" {
                # Playlist, numeric portion.
                incr next_pos  [GetInt $buffer 4]
                set  next_mhyp [expr {$pos + [GetInt $buffer 8]}]
                set  err_pos   $next_mhyp

                if {$index == $plist_num - 1} {
                    set err_state end
                } else {
                    set err_state plist_data
                }

                # Total number of tracks in playlist.
                set trk_num [GetInt $buffer 16]

                # Calculate default playlist name.
                if {([GetInt $buffer 20] & 255) == 1} {
                    set pl_name Master-PL
                    set pl_type master
                } else {
                    set pl_name Playlist
                    set pl_type custom
                }
                set pl_name [format "%s (%08x)" $pl_name $pos]

                # Expect text fields (mhod) or playlist entries (mhyp).
                set trk_idx 0
                set result  {}
                set state   plist_data_entry
            } "plist_data_entry mhod" {
                # Playlist, textual portion.
                incr next_pos [GetInt $buffer 8]

                # I assume the title mhod appears somewhere before the final
                # mhip entry--- if not, oh well, the defaults are fine. :^)

                if {[GetInt $buffer 12] == 1} {
                    # Playlist title.
                    set pl_name [GetUtf16 $buffer 40 [GetInt $buffer 28]]
                } else {
                    # Unknown mhod type (most likely 100).  Skip.
                }
            } "plist_data_entry mhip" {
                # Playlist, track entry.
                incr  next_pos [GetInt $buffer  8]
                lappend result [GetInt $buffer 24]

                incr trk_idx
                if {$trk_idx == $trk_num} {
                    # This was the last track in the playlist.  Skip any extra
                    # mhod's in this section.
                    set next_pos $next_mhyp
                    break
                }
            }

            default {
                # Basically I guess how to resynchronize on the next section.
                set err_pos   [GetInt $buffer 8]
                set err_state $state
                error "unexpected section \"$header\""
            }}

            } err] {
            0 {
                # TCL_OK: Success.  Go to the next block.
                set pos   $next_pos
            } 1 {
                # TCL_ERROR: Failure.  Attempt to resynchronize.
                set pos   $err_pos
                set state $err_state
                error     $err                
            } 3 {
                # TCL_BREAK: Success.  Return the completed record.
                set pos   $next_pos
                break
            }}
        }

        switch -- $goal {
        summary {
            set state end
            foreach which {Track Plist} {
                if {![info exists ${which}_pos]} {
                    set ${which}_num 0
                }
            }
            return [list summary        $path     \
                         track_count    $track_num\
                         playlist_count $plist_num]
        } track {
            incr index
            if {$index == $track_num} {
                set state end
            } else {
                set state track_data
            }
            return $result
        } playlist {
            incr index
            if {$index == $plist_num} {
                set state end
            } else {
                set state plist_data
            }
            return [list playlist $pl_name\
                         type     $pl_type\
                         tracks   $result]
        }}
    }

    # $db Get$type buffer offset ?length?
    #
    # Gets numeric or textual data from a buffer read from the database file.
    #
    # (Internal)
    #
    # $type - Type of data to get from the buffer.  Can be one of the following:
    #     Byte  -  8-bit signed value         (numeric) *UNIMPLEMENTED*
    #     Short - 16-bit signed value         (numeric) *UNIMPLEMENTED*
    #     Int   - 32-bit signed value         (numeric) (default)
    #     Long  - 64-bit signed value         (numeric) *UNIMPLEMENTED*
    #     Hdr   - four-byte header identifier (numeric)
    #     Bin   - binary data                 (textual) *UNIMPLEMENTED*
    #     Utf16 - UTF-16 data                 (textual)
    # $buffer - The byte buffer from which to extract data.
    # $offset - Offset relative to the start of $buffer at which to get data.
    # $length - Number of bytes to process.  Only valid with textual $type.
    #
    # return - The data extracted from the buffer.
    # error - "end of file" - Attempt to read past the end of the buffer.
    #
    # Full list of formats; use this to add support for the other types.  Just
    # use it as the second paramter to the foreach below.
    #
    #   byte  c       num     short s       num
    #   int   i       num     long  w       num
    #   hdr   a4      num     bin   binary  txt
    #   utf16 unicode txt
    foreach {type fmt mode} {
        int i num       hdr a4 num       utf16 unicode txt
    } {
        set name    Get[string totitle $type]
        set formals [list buffer offset]

        if {$mode eq "txt"} {
            lappend formals length

            set enc $fmt
            set fmt a\$length
        }

        set body [string map [list %FMT% $fmt] {
            if {[binary scan $buffer @${offset}%FMT% data] == 0} {
                error "end of file"
            }
        }]


        if {$mode eq "txt" && $enc ne "binary"} {
            append body [string map [list %ENC% $enc] {
                return [encoding convertfrom %ENC% $data]
            }]
        } else {
            append body {
                return $data
            }
        }

        proc $name $formals $body
    }
}

interp alias "" itunesdb "" itunesdb_impl %AUTO%

# vim: set ts=4 sts=4 sw=4 tw=80 et:
